/* sane - Scanner Access Now Easy.
   Copyright (C) 1996 David Mosberger-Tang and Andreas Beck
   This file is part of the SANE package.

   SANE is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   SANE is distributed in the hope that it will be useful, but WITHOUT
   ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
   or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
   License for more details.

   You should have received a copy of the GNU General Public License
   along with sane; see the file COPYING.  If not, write to the Free
   Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

   This file implements the backend-independent parts of SANE.  */
#include <stdio.h>

#include <sane.h>

int sanei_debug_level = 10;

extern const SANE_String_Const
sane_strstatus (SANE_Status status)
{
  static char buf[80];

  switch (status)
    {
    case SANE_STATUS_GOOD:
      return "Success";

    case SANE_STATUS_UNSUPPORTED:
      return "Operation not supported";

    case SANE_STATUS_CANCELLED:
      return "Operation was cancelled";

    case SANE_STATUS_DEVICE_BUSY:
      return "Device busy";

    case SANE_STATUS_AGAIN:
      return "No data available, try again";

    case SANE_STATUS_INVAL:
      return "Invalid argument";

    case SANE_STATUS_EOF:
      return "End of file reached";

    case SANE_STATUS_JAMMED:
      return "Document feeder jammed";

    case SANE_STATUS_NO_DOCS:
      return "Document feeder out of documents";

    case SANE_STATUS_COVER_OPEN:
      return "Scanner cover is open";

    case SANE_STATUS_IO_ERROR:
      return "Error during device I/O";

    case SANE_STATUS_NO_MEM:
      return "Out of memory";

    default:
      /* non-reentrant, but better than nothing */
      sprintf(buf, "Unknown SANE status code %d.", status);
      return buf;
    }
}


SANE_Status
sanei_constrain_value (const SANE_Option_Descriptor * opt, void * value,
		       SANE_Word * info)
{
  const SANE_String_Const * string_list;
  const SANE_Word * word_list;
  const SANE_Range * range;
  SANE_Word w, v;
  int i;

  switch (opt->constraint_type)
    {
    case SANE_CONSTRAINT_RANGE:
      w = *(SANE_Word *) value;
      range = opt->constraint.range;

      if (w < range->min || w > range->max)
	return SANE_STATUS_INVAL;		/* out of range */

      if (range->quant)
	{
	  v = (w - range->min + range->quant/2) / range->quant;
	  v = v * range->quant + range->min;
	  if (v != w)
	    {
	      *(SANE_Word *) value = v;
	      if (info)
		*info |= SANE_INFO_INEXACT;
	    }
	}
      break;

    case SANE_CONSTRAINT_WORD_LIST:
      w = *(SANE_Word *) value;
      word_list = opt->constraint.word_list;
      for (i = 1; w != word_list[i]; ++i)
	if (i >= word_list[0])
	  return SANE_STATUS_INVAL;
      break;

    case SANE_CONSTRAINT_STRING_LIST:
      string_list = opt->constraint.string_list;
      for (i = 0; strcmp(value, string_list[i]); ++i)
	if (!string_list[i + 1])
	  return SANE_STATUS_INVAL;
      break;

    default:
      break;
    }
  return SANE_STATUS_GOOD;
}
