/* SaneScan 0.5 -- plug-in for the GIMP image manipulation program
 * A scanner plug-in based on the SANE library.
 * Original program (C) 1996 Andreas Beck (becka@sunserver1.rz.uni-duesseldorf.de)
 * This version (C) 1996 Tristan Tarrant (tristan@galactica.it)
 *
 * You can contact the original "The GIMP" authors at gimp@xcf.berkeley.edu
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */


/* 
   BIG FAT WARNING !!!!!!!!!!!!!! 
   This program is still in development, and many things are missing or are buggy.
   Please bear with us...
   The program requires the following libraries:
   	libgtk.a, libgdk.a, libgimp.a, libglib.a, libsane.a (or their shared
	counterparts)
   Also gimp.h and gimp.c (from the plug-ins directory), need to be patched in the 
   following way (taken from Tom Bech's map-plane plug-in):

------------------------------ cut here -----------------------------------------
--- gimp/plug-ins/gimp.c        Sat Aug 24 03:25:10 1996
+++ sane/gimp.c Sun Nov 17 13:32:18 1996
@@ -218,6 +218,18 @@
   exit (0);
 }
 
+void
+gimp_noexit_quit (void)
+{
+  while (shared_memory)
+    {
+      shmdt (shared_memory->data);
+      shared_memory = shared_memory->next;
+    }
+
+  gimp_send (QUIT, 0, 0);
+}
+
 void*
 gimp_get_params ()
 {
--- gimp/plug-ins/gimp.h        Sat Aug 24 03:25:49 1996
+++ sane/gimp.h Sun Nov 17 13:31:28 1996
@@ -36,6 +36,7 @@
 void  gimp_install_load_save_handlers (GimpLoadSaveProc, GimpLoadSaveProc);
 void  gimp_main_loop (void);
 void  gimp_quit (void);
+void  gimp_noexit_quit (void);
 
 void* gimp_get_params (void);
 void  gimp_set_params (long, void *);
-------------------------------- cut here ------------------------------


   	Things which work:
   		Booleans,
		Groups,
		Ints and Fixed (Constraint range/string list),
		Strings,
		Buttons.
	Things which don't work:
		Ints and Fixed (Constraint word list/None)
		Preview... (maybe there is going to be a standard widget?)
	Things which are buggy:
		As soon as the info_label is updated the rulers screw up :-(



   */

/*
@(GIMP)         = <plug-in sanescan "Input/Scan">
@(GIMP_DEP)     = <gimpsane.c>
@(GIMP_OBJ)     = <gimpsane.o>
@(GIMP_LIB)     = <m c gimp>
@(GIMP_AUTHOR)  = <Andreas Beck & Tristan Tarrant>
@(GIMP_EMAIL)   = <becka@sunserver1.rz.uni-duesseldorf.de & tristan@galactica.it>
@(GIMP_DESC)    = <Interfaces to the SANE scanner library>
@(GIMP_VERSION) = <0.5>
@(GIMP_URL)     = <http://sunserver1.rz.uni-duesseldorf.de/~becka>
*/


#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include "gtk.h"
#ifdef __PLUGIN__
#include "gimp.h"
#endif
#include "sane.h"

/* Declare local functions. */
static void error_box(gchar *error);

#define MAXdlg			128
#define PREVIEW_PHYS_WIDTH	210
#define PREVIEW_PHYS_HEIGHT	297
#define PREVIEW_RATIO		1.5
#define PREVIEW_PIXEL_WIDTH	((int)(PREVIEW_PHYS_WIDTH*PREVIEW_RATIO))
#define PREVIEW_PIXEL_HEIGHT	((int)(PREVIEW_PHYS_HEIGHT*PREVIEW_RATIO))

typedef struct DlgElement {
    GtkWidget *widget;
    GtkData *data;
} DlgElement;

typedef struct MenuItem
{
    char *label;
    char  accelerator_key;
    int   accelerator_mods;
    GtkCallback callback;
    gpointer user_data;
    struct MenuItem *subitems;
    GtkWidget *widget;
} MenuItem;

GtkWidget *dialog;
GtkWidget *info_label;
DlgElement elements[MAXdlg];
const SANE_Device **devlist;
static SANE_Handle scan_handle;
gint seldev;		/* The selected device */
gint ndevs;		/* The number of available devices */


GtkWidget *build_menu (MenuItem *items, GtkAcceleratorTable *table)
{
    GtkWidget *menu;
    GtkWidget *menu_item;
    GtkData *data;
    
    menu = gtk_menu_new ();
    
    while (items->label) {
        if (items->label[0] == '-')
            menu_item = gtk_menu_item_new ();
        else {
            menu_item = gtk_menu_item_new_with_label (items->label);
            if (items->accelerator_key && table)
                gtk_widget_install_accelerator (menu_item,
                                                table,
                                                items->accelerator_key,
                                                items->accelerator_mods);
        }
        
        gtk_container_add (menu, menu_item);
        
        if (items->callback) {
            data = gtk_menu_item_get_state (menu_item);
            gtk_callback_add (data, items->callback, items->user_data);
        }
	
        if (items->subitems)
	    gtk_menu_item_set_submenu (menu_item, build_menu (items->subitems, table));
        
        gtk_widget_show (menu_item);
        items->widget = menu_item;
        
        items++;
    }
    
    return menu;
}


static void update_param(void)
{ 
    gchar buf[200];
    SANE_Parameters sp;
    
    if (sane_get_parameters(scan_handle,&sp)==SANE_STATUS_GOOD)
	sprintf(buf,"%d x %d : %d KB",sp.pixels_per_line,
		sp.lines,(sp.bytes_per_line*sp.lines)>>10);
    else
	sprintf(buf,"Invalid parameters.");
    if(info_label)
	gtk_label_set(info_label,buf);
}

static void showhide(gint item,gint onoff)
{ 
    if (onoff) {
	if(GTK_WIDGET_VISIBLE (elements[item].widget))
	    gtk_widget_hide(elements[item].widget);
    } else {
	if(!GTK_WIDGET_VISIBLE (elements[item].widget))
	    gtk_widget_show(elements[item].widget);
    }
}

static void update_complete_dialog(void)
{ 
    gint x;
    const SANE_Option_Descriptor *opt;

    for(x=0;(opt=sane_get_option_descriptor(scan_handle,x))!=NULL;x++) {
	switch(opt->type) {
	case SANE_TYPE_BOOL:
	    { 
		SANE_Word tog;

		sane_control_option(scan_handle,x,SANE_ACTION_GET_VALUE,&tog,NULL);
		gtk_check_button_set_state(elements[x].widget,tog);
		showhide(x,!SANE_OPTION_IS_ACTIVE(opt->cap));
		break;
	    } 
	case SANE_TYPE_INT:
	    switch(opt->constraint_type) {
	    case SANE_CONSTRAINT_RANGE: 
		{
		    SANE_Word tog;
		    GtkDataAdjustment *adjdata;
		    sane_control_option(scan_handle,x,SANE_ACTION_GET_VALUE,&tog,NULL);
		    adjdata = (GtkDataAdjustment *) elements[x].data;
		    adjdata->value = (double) (tog) * adjdata->step_increment;
		    gtk_data_notify (elements[x].data);
		    showhide(x,!SANE_OPTION_IS_ACTIVE(opt->cap));
		    break;
		} 
	    default:
		break;
	    }
	case SANE_TYPE_FIXED:
	    if(opt->constraint_type==SANE_CONSTRAINT_RANGE) { 
		SANE_Word tog;
		GtkDataAdjustment *adjdata;
		
		sane_control_option(scan_handle,x,SANE_ACTION_GET_VALUE,&tog,NULL);
		tog>>=SANE_FIXED_SCALE_SHIFT;
		adjdata = (GtkDataAdjustment *) elements[x].data;
		adjdata->value = (double) (tog) * adjdata->step_increment;
		gtk_data_notify (elements[x].data);
		showhide(x,!SANE_OPTION_IS_ACTIVE(opt->cap));
	    } 
	    break;
	case SANE_TYPE_STRING:
	    {
		char buf[512];

		sane_control_option(scan_handle,x,SANE_ACTION_GET_VALUE,buf,NULL);
		gtk_text_entry_set_text(elements[x].widget,buf);
		showhide(x,!SANE_OPTION_IS_ACTIVE(opt->cap));
		break;
	    } 
	default:
	    break;
	}
    }
}

void cancel_dialog_callback (GtkWidget *widget,
			     gpointer   client_data,
			     gpointer   call_data)
{
#ifdef __PLUGIN__
    gimp_noexit_quit();
#endif
    sane_exit();
    gtk_exit(0);
}

void scan_callback (GtkWidget *widget,
		    gpointer   client_data,
		    gpointer   call_data)
{
    SANE_Parameters parm;
    gchar *d;
    gint len,cnt,status;
#ifdef __PLUGIN__
    Image output=NULL;
#endif
    
    do {
  
	status=sane_start(scan_handle);
	if (status!=SANE_STATUS_GOOD) {
	    if (status!=SANE_STATUS_EOF) 
		error_box("Sane_start failed !");
	    break;
	}

	sane_get_parameters(scan_handle,&parm);
#ifdef __PLUGIN__
	if (parm.first_frame)	/* Open a new image */
	    output = gimp_new_image (0,parm.pixels_per_line,parm.lines,
				     (parm.format==SANE_FRAME_GREY)  ? GRAY_IMAGE : RGB_IMAGE );

	d = gimp_image_data (output);
	gimp_init_progress ("Receiving file ...");
#else
	d = (gchar *)malloc(5000);
#endif
	cnt = 0;
	switch(parm.format) {
	case SANE_FRAME_GREY:
	case SANE_FRAME_RGB:
	    while(sane_read(scan_handle,d,parm.bytes_per_line,&len)==SANE_STATUS_GOOD) {
		cnt+=len;
#ifdef __PLUGIN__
		d+=len;
		gimp_do_progress(cnt,parm.bytes_per_line*parm.lines);
#endif
	    }
	    break;
	default:	/* This client does not know how to handle other types of data */
	    break;
	}
#ifdef __PLUGIN__
	gimp_do_progress(1,1);
#else
	free(d);
#endif

    } while(1);

    sane_cancel(scan_handle);
#ifdef __PLUGIN__
    if (output) {
	gimp_display_image (output);
	gimp_update_image (output);
	gimp_free_image (output);
    }
    gimp_noexit_quit();
#endif
    sane_exit();
    gtk_exit(0);
}

void select_device_callback (GtkWidget *widget,
			     gpointer   client_data,
			     gpointer   call_data)
{
    seldev = (int)client_data;
}

void button_callback (GtkWidget *widget,
		      gpointer   client_data,
		      gpointer   call_data)
{
    SANE_Word info;
    
    sane_control_option(scan_handle,(int)client_data,SANE_ACTION_SET_VALUE,NULL,&info);
    if (info & SANE_INFO_RELOAD_OPTIONS)
	update_complete_dialog();
    if (info & SANE_INFO_RELOAD_PARAMS)
	update_param();
}

void info_callback (GtkWidget *widget,
		    gpointer   client_data,
		    gpointer   call_data)
{
}

void menu_callback (GtkWidget *widget,
		    gpointer   client_data,
		    gpointer   call_data)
{
    SANE_Word x, sel, info;

    x = (((int)client_data)&0xffff0000)>>16;
    sel = (((int)client_data)&0xffff);

    sane_control_option(scan_handle,x,SANE_ACTION_SET_VALUE,&sel,&info);
    if (info & SANE_INFO_RELOAD_OPTIONS)
	update_complete_dialog();
    if (info & SANE_INFO_RELOAD_PARAMS)
	update_param();
}

void close_dialog_callback (GtkWidget *widget,
			    gpointer   client_data,
			    gpointer   call_data)
{
    gtk_widget_hide(dialog);
    gtk_widget_destroy(dialog);
}

gint canvas_event (GtkWidget *canvas, GdkEvent *event)
{
    return TRUE;
}

void dummy_callback (GtkWidget *widget,
		     gpointer   client_data,
		     gpointer   call_data)
{
    return;
}

static gint toggle_callback_update (GtkObserver *observer,
				    GtkData *data)
{
    GtkDataInt *intdata;
    gint setting;
    SANE_Word info;
    
    intdata = (GtkDataInt *) data;
    if((intdata->value==GTK_STATE_ACTIVATED)||(intdata->value==GTK_STATE_DEACTIVATED)) {
	switch(intdata->value) {
	case GTK_STATE_ACTIVATED:
	    setting = 1;
	    break;
	case GTK_STATE_DEACTIVATED:
	    setting = 0;
	    break;
	}
	sane_control_option(scan_handle,(int)observer->user_data,
			    SANE_ACTION_SET_VALUE,&setting,&info);
	if (info & SANE_INFO_INEXACT) {
	    gtk_check_button_set_state(elements[(int)observer->user_data].widget,
				       setting);
	}
	if (info & SANE_INFO_RELOAD_OPTIONS)
	    update_complete_dialog();
	if (info & SANE_INFO_RELOAD_PARAMS)
	    update_param();
    }

    return 0;
}

static gint scale_callback_update (GtkObserver *observer,
				   GtkData *data)
{
    GtkDataAdjustment *adjdata;
    SANE_Word info, len;
    const SANE_Option_Descriptor *opt;

    opt=sane_get_option_descriptor(scan_handle,(int)observer->user_data);
    
    adjdata = (GtkDataAdjustment *)data;

    switch(opt->type) {
    case SANE_TYPE_INT:
	len = (int)adjdata->value;
	break;
    case SANE_TYPE_FIXED: 
	len = ((int)adjdata->value)<<SANE_FIXED_SCALE_SHIFT;
	break;
    default: 
	break;
    }

    sane_control_option(scan_handle,(int)observer->user_data,
			SANE_ACTION_SET_VALUE,&len,&info);
    
    if (info & SANE_INFO_INEXACT) {
	switch(opt->type) {
	case SANE_TYPE_INT:
	    break;
	case SANE_TYPE_FIXED: 
	    len>>=SANE_FIXED_SCALE_SHIFT;
	    break;
	default:
	    break;
	}
	adjdata->value = (double) (len) * adjdata->step_increment;
	gtk_data_notify (data);
    }
    if (info & SANE_INFO_RELOAD_OPTIONS) 
	update_complete_dialog();
    if (info & SANE_INFO_RELOAD_PARAMS) 
	update_param();
    return 0;
}

static void callback_disconnect (GtkObserver *observer,
				 GtkData *data)
{
}

static gint entry_callback (guint    keyval,
			    guint    state,
			    gpointer client_data)
{
    SANE_Word info;
    gchar buf[512];
    gchar *text;
    gint return_val;
    
    return_val = FALSE;
    text = gtk_text_entry_get_text (elements[(gint)client_data].widget);
    strcpy(buf,text);
    sane_control_option(scan_handle, (gint)client_data, SANE_ACTION_SET_VALUE, buf, &info);
    if (info & SANE_INFO_INEXACT)
	gtk_text_entry_set_text(elements[(gint)client_data].widget,buf);
    if (info & SANE_INFO_RELOAD_OPTIONS)
	update_complete_dialog();
    if (info & SANE_INFO_RELOAD_PARAMS)
	update_param();
    if (keyval == XK_Tab)
	return_val = TRUE;
    else if (keyval == XK_Return)
	return_val = TRUE;
    return return_val;
}

static void error_box(gchar *error)
{
    GtkWidget *main_vbox, *hbox, *label, *button, *frame;

    dialog = gtk_window_new("Error", GTK_WINDOW_TOPLEVEL);

    /* Create the main vbox */
    main_vbox = gtk_vbox_new(FALSE, 5);
    gtk_container_set_border_width(main_vbox, 0);
    gtk_container_add(dialog, main_vbox);
    gtk_widget_show(main_vbox);

    /* The top row with the message */
    hbox = gtk_hbox_new (FALSE, 5);
    gtk_container_set_border_width (hbox, 3);
    gtk_box_pack (main_vbox, hbox, TRUE, TRUE, 0, GTK_PACK_START);
    gtk_widget_show (hbox);

    /* The message */
    frame = gtk_frame_new(NULL);
    label = gtk_label_new(error);
    gtk_box_pack(hbox, frame, TRUE, TRUE, 0, GTK_PACK_START);
    gtk_container_add(frame,label);
    gtk_widget_show(label);
    gtk_widget_show(frame);

    /* The bottom row of buttons */
    hbox = gtk_hbox_new (FALSE, 5);
    gtk_container_set_border_width (hbox, 3);
    gtk_box_pack (main_vbox, hbox, TRUE, TRUE, 0, GTK_PACK_START);
    gtk_widget_show (hbox);

    /* The Ok button */
    button = gtk_push_button_new ();
    gtk_callback_add (gtk_button_get_state (button), cancel_dialog_callback, NULL);
    label = gtk_label_new ("Cancel");
    gtk_box_pack (hbox, button, TRUE, TRUE, 0, GTK_PACK_START);
    gtk_container_add (button, label);
    gtk_widget_show (label);
    gtk_widget_show (button);

    gtk_widget_show(dialog);
}


static void device_dialog ()
{
    GtkWidget *main_vbox, *vbox, *hbox, *button, *label;
    GtkWidget *table, *frame, *preview, *rule, *parent;
    GtkWidget *entry, *hbox1, *scale, *menu, *menu_list;
    MenuItem *menu_items;
    gchar buf[1000];
    GtkDataAdjustment *adjustment;
    GtkObserver *observer;
    gchar *devname;
    SANE_Status status;
    const SANE_Option_Descriptor *opt;
    SANE_Word sca;
    gint x, min, max, n, i;

    devname = devlist[seldev]->name;
    status = sane_open(devname,&scan_handle);
    if(status!=SANE_STATUS_GOOD) {
	error_box("Could not open selected device !");
	return;
    }

    dialog = gtk_window_new(devname, GTK_WINDOW_TOPLEVEL);
    
    /* Create the main vbox */
    main_vbox = gtk_vbox_new(FALSE, 5);
    gtk_container_set_border_width(main_vbox, 0);
    gtk_container_add(dialog, main_vbox);
    gtk_widget_show(main_vbox);

    /* Top portion of window - preview and options */
    hbox = gtk_hbox_new(FALSE, 5);
    gtk_container_set_border_width(hbox, 3);
    gtk_box_pack(main_vbox, hbox, TRUE, TRUE, 0, GTK_PACK_START);
    gtk_widget_show(hbox);

    /* Column 1 - the preview */
    vbox = gtk_vbox_new(FALSE, 5);
    gtk_container_set_border_width(vbox, 3);
    gtk_box_pack(hbox, vbox, TRUE, TRUE, 0, GTK_PACK_START);
    gtk_widget_show(vbox);

    /* The preview */
    table = gtk_table_new (2, 2, FALSE);
    gtk_table_set_col_spacing (table, 0, 1);
    gtk_table_set_row_spacing (table, 0, 1);
    gtk_container_set_border_width (table, 2);
    gtk_container_add (vbox, table);

    frame = gtk_frame_new(NULL);
    gtk_frame_set_type (frame, GTK_SHADOW_OUT);
    gtk_table_attach (table, frame, 0, 1, 1, 2, FALSE, FALSE, TRUE, 0,
		      FALSE, FALSE, TRUE, 0);
    gtk_widget_show (frame);

    adjustment = (GtkDataAdjustment *) gtk_data_adjustment_new (0, 0, PREVIEW_PHYS_WIDTH,
								1, 1, 1);
    rule = gtk_hruler_new (adjustment);
    gtk_ruler_set_metric (rule, PIXELS);
    gtk_table_attach (table, rule, 1, 2, 1, 2, TRUE, TRUE, TRUE, 0, FALSE,
		      FALSE, TRUE, 0);
    gtk_widget_show (rule);

    adjustment = (GtkDataAdjustment *) gtk_data_adjustment_new (0, 0, PREVIEW_PHYS_HEIGHT,
								1, 1, 1);
    rule = gtk_vruler_new (adjustment);
    gtk_ruler_set_metric (rule, PIXELS);
    gtk_table_attach (table, rule, 0, 1, 0, 1, FALSE, FALSE, TRUE, 0, 
		      TRUE, TRUE, TRUE, 0);
    gtk_widget_show (rule);

    preview = gtk_drawing_area_new (PREVIEW_PIXEL_WIDTH, PREVIEW_PIXEL_HEIGHT, canvas_event, 0);
    gtk_table_attach (table, preview, 1, 2, 0, 1, TRUE, TRUE, TRUE, 0, TRUE,
		      TRUE, TRUE, 0);
    gtk_widget_show (preview);
    gtk_widget_show (table);


    /* Column 2 - the options */
    vbox = gtk_vbox_new (FALSE, 5);
    gtk_container_set_border_width (vbox, 3);
    gtk_box_pack (hbox, vbox, TRUE, TRUE, 0, GTK_PACK_START);
    gtk_widget_show (vbox);
    
    parent = vbox;
    /* Build the options from the backend info */
    for(x=0;(opt=sane_get_option_descriptor(scan_handle,x))!=NULL;x++) {
	switch(opt->type) { 
	case SANE_TYPE_GROUP:
	    frame = gtk_frame_new (opt->name);
	    gtk_container_set_border_width(frame, 4);
	    gtk_frame_set_type(frame, GTK_SHADOW_ETCHED_IN);
	    gtk_box_pack(vbox, frame, FALSE, TRUE, 0, GTK_PACK_START);
	    gtk_widget_show (frame);
	    
	    parent = gtk_vbox_new(FALSE, 4);
	    gtk_container_set_border_width(parent, 2);
	    gtk_container_add(frame, parent);
	    gtk_widget_show(parent);
	    elements[x].widget = parent;
	    break;
	case SANE_TYPE_BOOL: /* We use a checkbutton for bool */
	    button = gtk_check_button_new ();

	    observer = (GtkObserver *)malloc(sizeof(GtkObserver));
	    observer->update = toggle_callback_update;
	    observer->disconnect = callback_disconnect;
	    observer->user_data = (void *) x;
	    gtk_data_attach(gtk_button_get_state(button),observer);
	    
	    sane_control_option(scan_handle,x,SANE_ACTION_GET_VALUE,&sca,NULL);
	    gtk_check_button_set_state(button,sca);

	    label = gtk_label_new (opt->name);
	    gtk_label_set_alignment (label, 0.0, 0.5);
	    gtk_box_pack (parent, button, FALSE, TRUE, 0, GTK_PACK_START);
	    gtk_container_add (button, label);
	    gtk_widget_show (label);
	    gtk_widget_show (button);
	    elements[x].widget = button;
	    break;
	case SANE_TYPE_INT:
	case SANE_TYPE_FIXED: { /* Scales for these */
	    switch(opt->constraint_type) { 
	    case SANE_CONSTRAINT_RANGE: /* Use a scale */
		sane_control_option(scan_handle,x,SANE_ACTION_GET_VALUE,&sca,NULL);
		switch(opt->type) {
		case SANE_TYPE_INT:
		    hbox1 = gtk_hbox_new (FALSE, 2);
		    gtk_container_set_border_width (hbox1, 0);
		    gtk_box_pack (parent, hbox1, FALSE, FALSE, 0, GTK_PACK_START);
		    label = gtk_label_new (opt->name);
		    gtk_box_pack (hbox1, label, FALSE, FALSE, 2, GTK_PACK_START);
		    min = opt->constraint.range->min;
		    max = opt->constraint.range->max;
		    adjustment = (GtkDataAdjustment*)
			gtk_data_adjustment_new (sca, min, max, 1, 1, 10);
		    scale = gtk_hscale_new (adjustment);
		    gtk_scale_set_value_pos (scale, GTK_POS_TOP);
		    gtk_scale_set_digits (scale, 0);
		    gtk_box_pack (hbox1, scale, TRUE, TRUE, 0, GTK_PACK_START);
		    gtk_widget_show (label);
		    gtk_widget_show (scale);
		    gtk_widget_show (hbox1);
		    elements[x].widget = scale;
		    elements[x].data = (GtkData *)adjustment;
		    break;
		case SANE_TYPE_FIXED:
		    hbox1 = gtk_hbox_new (FALSE, 2);
		    gtk_container_set_border_width (hbox1, 0);
		    gtk_box_pack (parent, hbox1, FALSE, FALSE, 0, GTK_PACK_START);
		    label = gtk_label_new (opt->name);
		    gtk_box_pack (hbox1, label, FALSE, FALSE, 2, GTK_PACK_START);
		    min = opt->constraint.range->min>>SANE_FIXED_SCALE_SHIFT;
		    max = opt->constraint.range->max>>SANE_FIXED_SCALE_SHIFT;
		    sca = sca>>SANE_FIXED_SCALE_SHIFT;
		    adjustment = (GtkDataAdjustment*) 
			gtk_data_adjustment_new (sca, min, max, 1, 1, 10);
		    
		    scale = gtk_hscale_new (adjustment);
		    gtk_scale_set_value_pos (scale, GTK_POS_TOP);
		    gtk_scale_set_digits (scale, 0);

		    observer = (GtkObserver *)malloc(sizeof(GtkObserver));
		    observer->update = scale_callback_update;
		    observer->disconnect = callback_disconnect;
		    observer->user_data = (void *) x;
		    gtk_data_attach ((GtkData *)adjustment, observer);

		    gtk_box_pack (hbox1, scale, TRUE, TRUE, 0, GTK_PACK_START);
		    gtk_widget_show (label);
		    gtk_widget_show (scale);
		    gtk_widget_show (hbox1);
		    elements[x].widget = scale;
		    elements[x].data = (GtkData *)adjustment;
		    break;
		default: 
		    fprintf(stderr,"This can't happen.\n");
		    break;
		}
		break;
	    case SANE_CONSTRAINT_WORD_LIST: /* Use a "list-select" widget */
		fprintf(stderr,"Can't do this yet\n");
		break;
	    case SANE_CONSTRAINT_STRING_LIST:
		hbox1 = gtk_hbox_new (FALSE, 2);
		gtk_container_set_border_width (hbox1, 0);
		gtk_box_pack (parent, hbox1, FALSE, FALSE, 0, GTK_PACK_START);
		label = gtk_label_new (opt->name);
		gtk_box_pack (hbox1, label, FALSE, FALSE, 2, GTK_PACK_START);
		menu = gtk_option_menu_new ();
		gtk_box_pack (hbox1, menu, FALSE, FALSE, 2, GTK_PACK_START);
		n = 0;
		while(opt->constraint.string_list[n])
		    ++n;
		menu_items = (MenuItem *)malloc((n+1)*sizeof(MenuItem));
		for(i=0;i<n;i++) {
		    menu_items[i].label = opt->constraint.string_list[i];
		    menu_items[i].accelerator_key = 0;
		    menu_items[i].accelerator_mods = 0;
		    menu_items[i].callback = menu_callback;
		    menu_items[i].user_data = (gpointer)(i+(x<<16));
		    menu_items[i].subitems = NULL;
		    menu_items[i].widget = NULL;
		}
		menu_items[n].label = NULL;
		menu_items[n].accelerator_key = 0;
		menu_items[n].accelerator_mods = 0;
		menu_items[n].callback = NULL;
		menu_items[n].user_data = NULL;
		menu_items[n].subitems = NULL;
		menu_items[n].widget = NULL;
		/* Fill in menu_items */
		menu_list = build_menu(menu_items,NULL);
		gtk_widget_show (label);
		gtk_widget_show (menu);
		gtk_widget_show (hbox1);
		gtk_option_menu_set_menu (menu, menu_list);
		elements[x].widget = menu;
		break;
	    case SANE_CONSTRAINT_NONE:/* Use a text widget*/
		break;
	    default:
		break;
	    }
	} 
	break;
	case SANE_TYPE_STRING:
	    entry = gtk_text_entry_new ();
	    gtk_text_entry_set_key_function (entry, entry_callback, (void *)x);
	    sane_control_option(scan_handle,x,SANE_ACTION_GET_VALUE,buf,NULL);
	    gtk_text_entry_set_text (entry, buf);
	    gtk_box_pack (parent, entry, FALSE, TRUE, 0, GTK_PACK_START);
	    gtk_widget_show (entry);
	    elements[x].widget = entry;
	    break;
	case SANE_TYPE_BUTTON:
	    button = gtk_push_button_new ();
	    gtk_callback_add (gtk_button_get_state (button), button_callback, (void *)x);
	    label = gtk_label_new (opt->name);
	    gtk_box_pack (parent, button, FALSE, TRUE, 0, GTK_PACK_START);
	    gtk_container_add (button, label);
	    gtk_widget_show (label);
	    gtk_widget_show (button);
	    elements[x].widget = button;
	    break;
	default:
	    break;
	}
    }

    /* The info row */
    hbox = gtk_hbox_new (FALSE, 5);
    gtk_container_set_border_width (hbox, 3);
    gtk_box_pack (main_vbox, hbox, TRUE, TRUE, 0, GTK_PACK_START);
    gtk_widget_show (hbox);

    info_label = gtk_label_new ("0 x 0 : 0 KB");
    gtk_label_set_alignment (info_label, 0.0, 0.5);
    gtk_box_pack (hbox, info_label, TRUE, TRUE, 0, GTK_PACK_START);
    gtk_widget_show (info_label);
    
    /* The bottom row of buttons */
    hbox = gtk_hbox_new (FALSE, 5);
    gtk_container_set_border_width (hbox, 3);
    gtk_box_pack (main_vbox, hbox, TRUE, TRUE, 0, GTK_PACK_START);
    gtk_widget_show (hbox);
    
    /* The Scan button */
    button = gtk_push_button_new ();
    gtk_callback_add (gtk_button_get_state (button), scan_callback, NULL);
    label = gtk_label_new ("Scan");
    gtk_box_pack (hbox, button, TRUE, TRUE, 0, GTK_PACK_START);
    gtk_container_add (button, label);
    gtk_widget_show (label);
    gtk_widget_show (button);

    /* The Preview button */
    button = gtk_push_button_new ();
    gtk_callback_add (gtk_button_get_state (button), cancel_dialog_callback, NULL);
    label = gtk_label_new ("Preview");
    gtk_box_pack (hbox, button, TRUE, TRUE, 0, GTK_PACK_START);
    gtk_container_add (button, label);
    gtk_widget_show (label);
    gtk_widget_show (button);
    
    /* The Cancel button */
    button = gtk_push_button_new ();
    gtk_callback_add (gtk_button_get_state (button), cancel_dialog_callback, NULL);
    label = gtk_label_new ("Cancel");
    gtk_box_pack (hbox, button, TRUE, TRUE, 0, GTK_PACK_START);
    gtk_container_add (button, label);
    gtk_widget_show (label);
    gtk_widget_show (button);
    
    gtk_widget_show(dialog);
}

void ok_choose_dialog_callback (GtkWidget *widget,
			     gpointer   client_data,
			     gpointer   call_data)
{
    gtk_widget_hide(dialog);
    gtk_widget_destroy(dialog);
    device_dialog ();
}

static void choose_device ()
{
    GtkWidget *main_vbox, *vbox, *hbox, *button, *label;
    GtkData *owner;
    const SANE_Device *adev;
    gint i;

   
    dialog = gtk_window_new("Select device", GTK_WINDOW_TOPLEVEL);
    
    main_vbox = gtk_vbox_new(FALSE,5);
    gtk_container_set_border_width (main_vbox, 0);
    gtk_container_add (dialog, main_vbox);
    gtk_widget_show (main_vbox);
    
    /* The list of drivers */
    vbox = gtk_vbox_new (FALSE, 5);
    gtk_container_set_border_width (vbox, 3);
    gtk_box_pack (main_vbox, vbox, TRUE, TRUE, 0, GTK_PACK_START);
    gtk_widget_show (vbox);
    
    /* The radio buttons */
    owner = NULL;
    for(i=0;i<ndevs;i++) {
	adev = devlist[i];
	
	button = gtk_radio_button_new (owner);
	gtk_callback_add(gtk_button_get_state(button),
			 select_device_callback,
			 (void *) i);
	label = gtk_label_new (adev->name);
	gtk_label_set_alignment (label, 0.0, 0.5);
	gtk_box_pack (vbox, button, TRUE, TRUE, 0, GTK_PACK_START);
	gtk_container_add (button, label);
	gtk_widget_show (label);
	gtk_widget_show (button);
	if(i==0)
	    owner = gtk_button_get_owner (button);
    }
    
    /* The bottom row of buttons */
    hbox = gtk_hbox_new (FALSE, 5);
    gtk_container_set_border_width (hbox, 3);
    gtk_box_pack (main_vbox, hbox, TRUE, TRUE, 0, GTK_PACK_START);
    gtk_widget_show (hbox);
    
    /* The Ok button */
    button = gtk_push_button_new ();
    gtk_callback_add (gtk_button_get_state (button), ok_choose_dialog_callback, NULL);
    label = gtk_label_new ("Ok");
    gtk_box_pack (hbox, button, TRUE, TRUE, 0, GTK_PACK_START);
    gtk_container_add (button, label);
    gtk_widget_show (label);
    gtk_widget_show (button);
    
    /* The Cancel button */
    button = gtk_push_button_new ();
    gtk_callback_add (gtk_button_get_state (button), cancel_dialog_callback, NULL);
    label = gtk_label_new ("Cancel");
    gtk_box_pack (hbox, button, TRUE, TRUE, 0, GTK_PACK_START);
    gtk_container_add (button, label);
    gtk_widget_show (label);
    gtk_widget_show (button);
    
    gtk_widget_show (dialog);
}

int main (int argc, char **argv)
{
    const SANE_Device *adev;
    info_label = NULL;
#ifdef __PLUGIN__
    if(gimp_init (argc, argv)) {
#endif
	gdk_set_debug_level (0);
	gdk_set_show_events (0);
	gtk_init (&argc, &argv);
	
	sane_init(NULL);
	
	sane_get_devices(&devlist);
	
	ndevs = 0;
	adev = devlist[0];
	while(adev)
	    adev = devlist[++ndevs];
	
	if(ndevs==1) {
	    seldev = 0;
	    device_dialog();
	} else
	    choose_device();
	
	gtk_main();
	
	sane_exit();
#ifdef __PLUGIN__
	gimp_noexit_quit();
#endif
	gtk_exit(0);
#ifdef __PLUGIN__    
    } else 
	fprintf(stderr,"This module can only be invoked from within the GIMP\n");
#endif
    return 0;
}
