/* sane - Scanner Access Now Easy.
   Copyright (C) 1996 David Mosberger-Tang and Andreas Beck
   This file is part of the SANE package.

   SANE is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2 of the License, or (at your
   option) any later version.

   SANE is distributed in the hope that it will be useful, but WITHOUT
   ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.

   You should have received a copy of the GNU General Public License
   along with sane; see the file COPYING.  If not, write to the Free
   Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

   This file declares common option names, titles, and descriptions.  A
   backend is not limited to these options but for the sake of
   consistency it's better to use options declared here when appropriate.
 */
#ifndef saneopts_h
#define saneopts_h

/* This _must_ be the first option (index 0): */
#define SANE_NAME_NUM_OPTIONS		""	/* never settable */
#define SANE_NAME_BIT_DEPTH		"depth"
#define SANE_NAME_SCAN_MODE		"mode"
#define SANE_NAME_SCAN_SPEED		"speed"
#define SANE_NAME_SCAN_SOURCE		"source"
#define SANE_NAME_BACKTRACK		"backtrack"
/* Most user-interfaces will let the user specify the scan area as the
   top-left corner and the width/height of the scan area.  The reason
   the backend interface uses the top-left/bottom-right corner is so
   that the scan area values can be properly constraint independent of
   any other option value.  */
#define SANE_NAME_SCAN_TL_X		"tl-x"
#define SANE_NAME_SCAN_TL_Y		"tl-y"
#define SANE_NAME_SCAN_BR_X		"br-x"
#define SANE_NAME_SCAN_BR_Y		"br-y"
#define SANE_NAME_SCAN_RESOLUTION	"resolution"
#define SANE_NAME_GAMMA_LENGTH		"gamma-length"
#define SANE_NAME_GAMMA_VECTOR		"gamma-table"
#define SANE_NAME_GAMMA_VECTOR_R	"red-gamma-table"
#define SANE_NAME_GAMMA_VECTOR_G	"green-gamma-table"
#define SANE_NAME_GAMMA_VECTOR_B	"blue-gamma-table"
#define SANE_NAME_BRIGHTNESS		"brightness"
#define SANE_NAME_CONTRAST		"contrast"
#define SANE_NAME_GRAIN_SIZE		"grain"
#define SANE_NAME_HALFTONE		"halftoning"
#define SANE_NAME_BLACKLEVEL		"blacklevel"
#define SANE_NAME_WHITELEVEL		"whitelevel"
#define SANE_NAME_HUE			"hue"
#define SANE_NAME_SATURATION		"saturation"
#define SANE_NAME_FILE			"filename"
#define SANE_NAME_HALFTONE_DIMENSION	"halftone-size"
#define SANE_NAME_HALFTONE_PATTERN	"halftone-pattern"

#define SANE_TITLE_NUM_OPTIONS		"Number of options"
#define SANE_TITLE_BIT_DEPTH		"Bit depth"
#define SANE_TITLE_SCAN_MODE		"Scan mode"
#define SANE_TITLE_SCAN_SPEED		"Scan speed"
#define SANE_TITLE_SCAN_SOURCE		"Scan source"
#define SANE_TITLE_BACKTRACK		"Force backtracking"
#define SANE_TITLE_SCAN_TL_X		"Top-left x position of scan area"
#define SANE_TITLE_SCAN_TL_Y		"Top-left y position of scan area"
#define SANE_TITLE_SCAN_BR_X		"Bottom-right x position of scan area"
#define SANE_TITLE_SCAN_BR_Y		"Bottom-right y position of scan area"
#define SANE_TITLE_SCAN_RESOLUTION	"Scan resolution"
#define SANE_TITLE_GAMMA_LENGTH		"Gamma table length"
#define SANE_TITLE_GAMMA_VECTOR		"Gamma table"
#define SANE_TITLE_GAMMA_VECTOR_R	"Red gamma table"
#define SANE_TITLE_GAMMA_VECTOR_G	"Green gamma table"
#define SANE_TITLE_GAMMA_VECTOR_B	"Blue gamma table"
#define SANE_TITLE_BRIGHTNESS		"Brightness"
#define SANE_TITLE_CONTRAST		"Contrast"
#define SANE_TITLE_GRAIN_SIZE		"Grain size"
#define SANE_TITLE_HALFTONE		"Halftoning"
#define SANE_TITLE_BLACKLEVEL		"Blacklevel"
#define SANE_TITLE_WHITELEVEL		"Whitelevel"
#define SANE_TITLE_HUE			"Hue"
#define SANE_TITLE_SATURATION		"Saturation"
#define SANE_TITLE_FILE			"Filename"
#define SANE_TITLE_HALFTONE_DIMENSION	"Halftone pattern size"
#define SANE_TITLE_HALFTONE_PATTERN	"Halftone pattern"

/* Descriptive/help strings for above options: */
#define SANE_DESC_NUM_OPTIONS \
"Read-only option that specifies how many options a specific devices " \
"supports."

#define SANE_DESC_BIT_DEPTH \
"Number of bits per sample, typical values are 1 for \"line-art\" and 8 " \
"for multibit scans."

#define SANE_DESC_SCAN_MODE \
"Selects the scan mode (e.g., lineart,monochrome, or color)."

#define SANE_DESC_SCAN_SPEED \
"Determines the speed at which the scan proceeds."

#define SANE_DESC_SCAN_SOURCE \
"Selects the scan source (such as a document-feeder)."

#define SANE_DESC_BACKTRACK \
"Controls whether backtracking is forced."

#define SANE_DESC_SCAN_TL_X \
"Top-left x position of scan area."

#define SANE_DESC_SCAN_TL_Y \
"Top-left y position of scan area."

#define SANE_DESC_SCAN_BR_X \
"Bottom-right x position of scan area."

#define SANE_DESC_SCAN_BR_Y \
"Bottom-right y position of scan area."

#define SANE_DESC_SCAN_RESOLUTION \
"Sets the resolution of the scanned image."

#define SANE_DESC_GAMMA_LENGTH \
"Length of gamma-correction table.  A length of 0 implies that a built" \
"in gamma correction is to be used (e.g., a simple linear table)."

#define SANE_DESC_GAMMA_VECTOR \
"Gamma-correction table.  In color mode this option equally affects the " \
"red, green, and blue channels simultaneously (i.e., it is an intensity " \
"gamma table)."

#define SANE_DESC_GAMMA_VECTOR_R \
"Gamma-correction table for the red band."

#define SANE_DESC_GAMMA_VECTOR_G \
"Gamma-correction table for the green band."

#define SANE_DESC_GAMMA_VECTOR_B \
"Gamma-correction table for the blue band."

#define SANE_DESC_BRIGHTNESS \
"Controls the brightness of the acquired image."

#define SANE_DESC_CONTRAST \
"Controls the contrast of the acquired image."

#define SANE_DESC_GRAIN_SIZE \
"Selects the \"graininess\" of the acquired image.  Smaller values " \
"result in sharper images."

#define SANE_DESC_HALFTONE \
"Selects whether the acquired image should be halftoned (dithered)."

#define SANE_DESC_BLACKLEVEL \
"Selects what radiance level should be considered \"black\"."

#define SANE_DESC_WHITELEVEL \
"Selects what radiance level should be considered \"white\"."

#define SANE_DESC_HUE \
"Controls the \"hue\" (blue-level) of the acquired image."

#define SANE_DESC_SATURATION \
"The saturation level controls the amount of \"blooming\" that occurs when " \
"acquiring an image with a camera. Larger values cause more blooming."

#define SANE_DESC_FILE "The filename of the image to be loaded."

#define SANE_DESC_HALFTONE_DIMENSION \
"Sets the size of the halftoning (dithering) pattern used when scanning " \
"halftoned images."

#define SANE_DESC_HALFTONE_PATTERN \
"Defines the halftoning (dithering) pattern for scanning halftoned images."

#endif /* saneopts_h */
